import pandas as pd
import numpy as np
import sys
from scipy.optimize import newton
from tqdm import tqdm

class npv_calc():
    
    def __init__(self,cashflow:np.array):
        self._cashflow = cashflow

    def constraint(self,x):
        return x


    def get_npv(self,discount_rate):
        #if np.sum(self._cashflow) < 0:
        #    l_npv = -1
        #elif discount_rate < 0:
        #    l_npv = 99999
        #else:
        l_npv = np.sum(self._cashflow/ (1+discount_rate)**np.arange(0, len(self._cashflow)))

        return l_npv
    
    def get_irr(self, iterations=50):
        guess = .1/12

        constraint_bounds = ((1e-6, .16),) 

        if np.sum(self._cashflow) < 0:
            result = 0
        else:
            result = newton(self.get_npv, guess, maxiter=iterations)

        try:
            result = np.power(1+result,12)-1
        except Exception as e:
            print(result)
            raise e

        return result

class well_econ:

    def __init__(self,verbose=False):
        #Constants
        self.STAT_FILE = sys.stdout
        self.OIL_COL = "OIL"
        self.GAS_COL = 'GAS'
        self.WATER_COL = 'WATER'

        self._verbose = verbose

        #Settable
        self._flowstreams = None
        self._header_data = None

        self._flowstream_uwi_col = None
        self._flowstream_t_index = None
        self._header_uwi_col = None

        self._royalty = None

        self._opc_t = None
        self._opc_oil = None
        self._opc_gas = None
        self._opc_water = None

        self._scale_capex = False
        self._scale_column = None
        self._capex_val = None
        self._capex_col = None # Careful, will override any prior settings

        self._atx = None
        self._sev_gas = None
        self._sev_oil = None

        self._oil_pri = None
        self._gas_pri = None
        
        self._discount_rate = None

        self._breakeven_phase = None

        # optional variables
        self._royalty_col = None

        self._wi_col = None
        self._nri_col = None

        self._gas_shrink = 0 
        self._ngl_yield = 0 # input as b/M post shrink
        self._ngl_price_fraction = 0
        self._scale_forecast = False
        self._scale_base = 5280
        self._oil_diff = 0
        self._gas_diff = 0

        self._spud_to_online = None
        self._t_start_column = None

        #Get only
        self._indicators = None

    @property
    def nri_col(self):
        return self._nri_col


    @nri_col.setter
    def nri_col(self,value):
        self._nri_col = value

    @property
    def wi_col(self):
        return self._wi_col


    @wi_col.setter
    def wi_col(self,value):
        self._wi_col = value

    @property
    def capex_col(self):
        return self._capex_col


    @capex_col.setter
    def capex_col(self,value):
        self._capex_col = value

    @property
    def gas_diff(self):
        return self._gas_diff


    @gas_diff.setter
    def gas_diff(self,value):
        self._gas_diff = value

    @property
    def oil_diff(self):
        return self._oil_diff


    @oil_diff.setter
    def oil_diff(self,value):
        self._oil_diff = value

    @property
    def t_start_column(self):
        return self._t_start_column


    @t_start_column.setter
    def t_start_column(self,value):
        self._t_start_column = value

    @property
    def scale_forecast(self):
        return self._scale_forecast


    @scale_forecast.setter
    def scale_forecast(self,value):
        self._scale_forecast = value

    @property
    def spud_to_online(self):
        return self._spud_to_online


    @spud_to_online.setter
    def spud_to_online(self,value):
        self._spud_to_online = value

    @property
    def ngl_price_fraction(self):
        return self._ngl_price_fraction


    @ngl_price_fraction.setter
    def ngl_price_fraction(self,value):
        self._ngl_price_fraction = value

    @property
    def ngl_yield(self):
        return self._ngl_yield


    @ngl_yield.setter
    def ngl_yield(self,value):
        self._ngl_yield = value

    @property
    def gas_shrink(self):
        return self._gas_shrink


    @gas_shrink.setter
    def gas_shrink(self,value):
        self._gas_shrink = value

    @property
    def gas_shrink(self):
        return self._gas_shrink


    @gas_shrink.setter
    def gas_shrink(self,value):
        self._gas_shrink = value

    @property
    def royalty_col(self):
        return self._royalty_col


    @royalty_col.setter
    def royalty_col(self,value):
        self._royalty_col = value
        

    @property
    def flowstreams(self):
        return self._flowstreams


    @flowstreams.setter
    def flowstreams(self,value):
        self._flowstreams = value

    @property
    def flowstream_uwi_col(self):
        return self._flowstream_uwi_col


    @flowstream_uwi_col.setter
    def flowstream_uwi_col(self,value):
        self._flowstream_uwi_col = value

    @property
    def flowstream_t_index(self):
        return self._flowstream_t_index


    @flowstream_t_index.setter
    def flowstream_t_index(self,value):
        self._flowstream_t_index = value

    @property
    def header_uwi_col(self):
        return self._header_uwi_col


    @header_uwi_col.setter
    def header_uwi_col(self,value):
        self._header_uwi_col = value

    @property
    def header_data(self):
        return self._header_data


    @header_data.setter
    def header_data(self,value):
        self._header_data = value

    @property
    def royalty(self):
        return self._royalty


    @royalty.setter
    def royalty(self,value):
        self._royalty = value

    @property
    def opc_t(self):
        return self._opc_t


    @opc_t.setter
    def opc_t(self,value):
        self._opc_t = value

    @property
    def opc_oil(self):
        return self._opc_oil


    @opc_oil.setter
    def opc_oil(self,value):
        self._opc_oil = value

    @property
    def opc_gas(self):
        return self._opc_gas


    @opc_gas.setter
    def opc_gas(self,value):
        self._opc_gas = value

    @property
    def opc_water(self):
        return self._opc_water


    @opc_water.setter
    def opc_water(self,value):
        self._opc_water = value

    @property
    def scale_capex(self):
        return self._scale_capex


    @scale_capex.setter
    def scale_capex(self,value):
        self._scale_capex = value

    @property
    def scale_column(self):
        return self._scale_column


    @scale_column.setter
    def scale_column(self,value):
        self._scale_column = value

    @property
    def capex_val(self):
        return self._capex_val


    @capex_val.setter
    def capex_val(self,value):
        self._capex_val = value

    @property
    def atx(self):
        return self._atx


    @atx.setter
    def atx(self,value):
        self._atx = value

    @property
    def sev_gas(self):
        return self._sev_gas


    @sev_gas.setter
    def sev_gas(self,value):
        self._sev_gas = value

    @property
    def sev_oil(self):
        return self._sev_oil


    @sev_oil.setter
    def sev_oil(self,value):
        self._sev_oil = value

    @property
    def oil_pri(self):
        return self._oil_pri


    @oil_pri.setter
    def oil_pri(self,value):
        self._oil_pri = value

    @property
    def gas_pri(self):
        return self._gas_pri


    @gas_pri.setter
    def gas_pri(self,value):
        self._gas_pri = value

    @property
    def discount_rate(self):
        return self._discount_rate


    @discount_rate.setter
    def discount_rate(self,value):
        self._discount_rate = value

    @property
    def indicators(self):
        return self._indicators

    @property
    def breakeven_phase(self):
        return self._breakeven_phase

    @breakeven_phase.setter
    def breakeven_phase(self,value):
        self._breakeven_phase = value

   

    def generate_oil_price(self,times):
        oil_price = []
        if isinstance(self._oil_pri, list):
            if len(self._oil_pri) >= len(times):
                oil_price = self._oil_pri[0:len(times)]
            else:
                last_pri = self._oil_pri[-1]
                num_to_add = len(times)-len(self._oil_pri)
                add_list = [last_pri for i in range(num_to_add)]
                oil_price = self._oil_pri
                oil_price.extend(add_list)
        else:
            oil_price = [self._oil_pri for i in range(len(times))]

        return np.array(oil_price)+self.oil_diff

    def generate_gas_price(self,times):
        gas_price = []
        if isinstance(self._gas_pri, list):
            if len(self._gas_pri) >= len(times):
                gas_price = self._gas_pri[0:len(times)]
            else:
                last_pri = self._gas_pri[-1]
                num_to_add = len(times)-len(self._gas_pri)
                add_list = [last_pri for i in range(num_to_add)]
                gas_price = self._gas_pri
                gas_price.extend(add_list)
        else:
            gas_price = [self._gas_pri for i in range(len(times))]

        return np.array(gas_price)+self.gas_diff

    def generate_capex(self,times,well):
        l_capex = np.zeros(times)

        if self._t_start_column:
            #capex_point = self._header_data[self._header_data[self._header_uwi_col]==well].iloc[0][self._t_start_column]
            capex_point=0
        else:
            capex_point=0

        if self._capex_col:
            capex_val = self._header_data[self._header_data[self._header_uwi_col]==well].iloc[0][self._capex_col]
            l_capex[capex_point] = capex_val

        elif self._scale_capex:

            scale_val = self._header_data[self._header_data[self._header_uwi_col]==well].iloc[0][self._scale_column]

            l_capex[capex_point] = self.capex_val*scale_val
        else:
            l_capex[capex_point] = self._capex_val

        return l_capex

    def zero_below(self,df:pd.DataFrame,i_max:int, cols:list):
        for col in cols:
            df[col] = np.where(
                df['T_INDEX'] <= i_max,
                df[col],
                0
            )

        return df

    def well_flowstream(self,input_well):

        l_flow = self._flowstreams[self._flowstreams[self._flowstream_uwi_col]==input_well].reset_index(drop=True)

        if self._scale_forecast:
            scale_val = self._header_data[self._header_data[self._header_uwi_col]==input_well].iloc[0][self._scale_column]
            l_flow[self.OIL_COL] = l_flow[self.OIL_COL]*scale_val/self._scale_base
            l_flow[self.GAS_COL] = l_flow[self.GAS_COL]*scale_val/self._scale_base

        if self._t_start_column:
            start_val = self._header_data[self._header_data[self._header_uwi_col]==input_well].iloc[0][self._t_start_column]
            l_uid = l_flow[self._flowstream_uwi_col].iloc[0]
            l_major = l_flow['MAJOR'].iloc[0]
            l_t_index = pd.Series(range(1,start_val+1,1))

            # Create a DataFrame of zeros
            start_df = pd.DataFrame(0, index=range(start_val), columns=l_flow.columns)
            start_df[self._flowstream_uwi_col] = l_uid
            start_df['MAJOR'] = l_major
            start_df[self._flowstream_t_index] = l_t_index

            # Concatenate the zeros DataFrame with the original DataFrame
            l_flow[self._flowstream_t_index] += start_val
            #l_flow = pd.concat([zeros_df, l_flow]).reset_index(drop=True)
        else:
            start_val = 0
            start_df = pd.DataFrame([])
        
        if self._spud_to_online:
            n = self._spud_to_online  # Number of rows to insert

            l_uid = l_flow[self._flowstream_uwi_col].iloc[0]
            l_major = l_flow['MAJOR'].iloc[0]
            l_t_index = pd.Series(range(start_val+1,start_val+n+1,1))

            # Create a DataFrame of zeros
            zeros_df = pd.DataFrame(0, index=range(n), columns=l_flow.columns)
            zeros_df[self._flowstream_uwi_col] = l_uid
            zeros_df['MAJOR'] = l_major
            zeros_df[self._flowstream_t_index] = l_t_index

            if not start_df.empty:
                zeros_df = pd.concat([start_df,zeros_df])

            # Concatenate the zeros DataFrame with the original DataFrame
            l_flow[self._flowstream_t_index] += n
            l_flow = pd.concat([zeros_df, l_flow]).reset_index(drop=True)
        else:
            n = 0

        
        l_flow['gas_sold'] = l_flow[self.GAS_COL]*(1-self._gas_shrink)
        l_flow['ngl_volume'] = l_flow['gas_sold']*self._ngl_yield
        t_series = np.array(range(len(l_flow)))

        l_flow['oil_price'] = self.generate_oil_price(t_series)
        l_flow['gas_price'] = self.generate_gas_price(t_series)
        l_flow['ngl_price'] = l_flow['oil_price'] * self._ngl_price_fraction

        l_flow['revenue'] = (
            l_flow[self.OIL_COL]*l_flow['oil_price']
            +l_flow['gas_sold']*l_flow['gas_price']
            +l_flow['ngl_volume']*l_flow['ngl_price']
        )

        if self._wi_col and self._nri_col:
            l_nri = self._header_data[self._header_data[self._header_uwi_col]==input_well].iloc[0][self._nri_col]
            l_wi = self._header_data[self._header_data[self._header_uwi_col]==input_well].iloc[0][self._wi_col]
            l_royalty = 1-l_nri/l_wi
            l_flow['royalty'] = l_flow['revenue']*l_royalty
        elif self._royalty_col:
            l_royalty = self._header_data[self._header_data[self._header_uwi_col]==input_well].iloc[0][self._royalty_col]
            l_flow['royalty'] = l_flow['revenue']*l_royalty
            l_wi = 1
        else:
            l_royalty = self._royalty
            l_flow['royalty'] = l_flow['revenue']*self._royalty
            l_wi = 1

        l_flow['net_oil'] = l_flow[self.OIL_COL]*l_royalty
        l_flow['net_gas'] = l_flow['gas_sold']*l_royalty
        l_flow['net_ngl'] = l_flow['ngl_volume']*l_royalty

        l_flow['expense'] = (
            self._opc_t
            + self._opc_gas*l_flow['gas_sold']
            + self._opc_oil*l_flow[self.OIL_COL]
            + self._opc_water*l_flow[self.WATER_COL]
        )

        l_flow['expense'] = np.where(
            l_flow[self._flowstream_t_index] < start_val+n,
            0,
            l_flow['expense']
        )

        l_flow['taxes'] = (
            self._atx*l_flow['revenue']
            + self._sev_gas*l_flow[self.GAS_COL]*l_flow['gas_price']
            + self._sev_oil*l_flow[self.OIL_COL]*l_flow['oil_price']
        )*(1-l_royalty)

        l_flow['capex'] = self.generate_capex(len(l_flow),input_well)

        l_flow['cf'] = (
            l_flow['revenue']
            - l_flow['royalty']
            - l_flow['expense']
            - l_flow['taxes']
            - l_flow['capex']
        )


        l_flow['dcf'] = (l_flow['cf'].to_numpy() / (1+self._discount_rate)**np.arange(0, len(l_flow['cf'].to_numpy())))
        

        try:
            cf_idx = np.argwhere(l_flow['dcf'].to_numpy()>0)
        except:
            cf_idx=[]

        if len(cf_idx) > 0:
            last_cf = np.max(np.argwhere(l_flow['dcf'].to_numpy()>0))
        else:
            last_cf = 0

        zero_cols = [
            self.OIL_COL,
            self.GAS_COL,
            'gas_sold',
            self.WATER_COL,
            'ngl_volume',
            'net_oil',
            'net_gas',
            'net_ngl',
            'revenue',
            'royalty',
            'expense',
            'taxes',
            'capex',
            'cf',
            'dcf'
        ]

        l_flow = self.zero_below(l_flow,last_cf,zero_cols)

        l_flow[['wi_net_oil',
            'wi_net_gas',
            'wi_net_ngl',
            'wi_revenue',
            'wi_royalty',
            'wi_expense',
            'wi_taxes',
            'wi_capex',
            'wi_cf',
            'wi_dcf']] = l_flow[['net_oil',
            'net_gas',
            'net_ngl',
            'revenue',
            'royalty',
            'expense',
            'taxes',
            'capex',
            'cf',
            'dcf']].mul(l_wi)

        # calculate WI vales

        return l_flow


    def generate_indicators(self):

        ind_dict = {
            'UWI':[],
            'EURO':[],
            'EURG':[],
            'EURW':[],
            'REVENUE':[],
            'ROYALTY':[],
            'OPEX':[],
            'TAXES':[],
            'CAPEX':[],
            'FCF':[],
            'DCF':[],
            'IRR':[],
            'ROI':[],
            'PAYOUT':[],
            'BREAKEVEN':[],
            'BREAKEVEN_PHASE':[]
        }

        unique_wells = self._flowstreams[self._flowstream_uwi_col].unique()

        iterable = tqdm(unique_wells) if self._verbose else unique_wells

        for w in iterable:
            
            l_flow = self.well_flowstream(w)
            #l_flow.to_csv(f'tests/{w}.csv')
            
            dc_rev = (l_flow['revenue'].to_numpy() / (1+self._discount_rate)**np.arange(0, len(l_flow['revenue'].to_numpy())))

            if self._breakeven_phase is None:
                if np.sum(l_flow[self.OIL_COL]) > 0:
                    if np.sum(l_flow[self.GAS_COL])/np.sum(l_flow[self.OIL_COL])> 3.2:
                        be_major = 'GAS'
                        break_even =(np.sum(dc_rev)- np.sum(l_flow['dcf']))/np.sum(l_flow[self.GAS_COL])
                    else:    
                        be_major = 'OIL'
                        break_even =(np.sum(dc_rev)- np.sum(l_flow['dcf']))/np.sum(l_flow[self.OIL_COL])
                else:
                    be_major = 'GAS'
                    break_even =(np.sum(dc_rev)- np.sum(l_flow['dcf']))/np.sum(l_flow[self.GAS_COL])
            else:
                if self._breakeven_phase == "GAS":
                    be_major = 'GAS'
                    break_even =(np.sum(dc_rev)- np.sum(l_flow['dcf']))/np.sum(l_flow[self.GAS_COL])
                else:
                    be_major = 'OIL'
                    break_even =(np.sum(dc_rev)- np.sum(l_flow['dcf']))/np.sum(l_flow[self.OIL_COL])

            l_flow['cum_cf'] = l_flow['cf'].cumsum()
            positive_index = (l_flow['cum_cf'] > 0).idxmax()


            ind_dict['UWI'].append(w)
            ind_dict['EURO'].append(np.sum(l_flow[self.OIL_COL]))
            ind_dict['EURG'].append(np.sum(l_flow[self.GAS_COL]))
            ind_dict['EURW'].append(np.sum(l_flow[self.WATER_COL]))
            ind_dict['REVENUE'].append(np.sum(l_flow['revenue']))
            ind_dict['ROYALTY'].append(np.sum(l_flow['royalty']))
            ind_dict['OPEX'].append(np.sum(l_flow['expense']))
            ind_dict['TAXES'].append(np.sum(l_flow['taxes']))
            ind_dict['CAPEX'].append(np.sum(l_flow['capex']))
            ind_dict['FCF'].append(np.sum(l_flow['cf']))
            ind_dict['DCF'].append(np.sum(l_flow['dcf']))
            ind_dict['BREAKEVEN'].append(break_even)
            ind_dict['BREAKEVEN_PHASE'].append(be_major)
            ind_dict['PAYOUT'].append(positive_index)

            #if np.sum(cf_array) > 0:
            try:
                l_npv = npv_calc(l_flow['cf'].to_numpy())
                #print(l_npv.get_npv(0))
                ind_dict['IRR'].append(l_npv.get_irr())
            except:
                ind_dict['IRR'].append(0)

            ind_dict['ROI'].append(np.sum(l_flow['cf'])/np.sum(l_flow['capex'])+1)
            #else:
            #    ind_dict['IRR'].append(0)


        self._indicators = pd.DataFrame(ind_dict)

    def generate_cashflow(self):
        r_df = pd.DataFrame([])
        unique_wells = self._flowstreams[self._flowstream_uwi_col].unique()

        iterable = tqdm(unique_wells) if self._verbose else unique_wells

        for w in iterable:
            l_flow = self.well_flowstream(w)

            if r_df.empty:
                r_df = l_flow
            else:
                r_df = pd.concat([r_df,l_flow])

        return r_df


from re import S
import pandas as pd
import numpy as np
import sys
import statsmodels.api as sm
from scipy.signal import argrelextrema
from scipy.optimize import curve_fit, fsolve
from dateutil.relativedelta import *
import time
import math

import warnings
from tqdm import tqdm
warnings.simplefilter("ignore")

class decline_solver():

    def __init__(self, qi=None, qf=None, de=None, dmin=None, b=None, eur=None, t_max=None):

        
        self.qi = qi
        self.qf = qf
        self.de = de
        self.dmin = dmin
        self.b = b
        self.eur = eur
        self.t_max = t_max

        self.l_qf = qf
        self.l_t_max = t_max
        self.delta = 0
        
        self.variables_to_solve = []

        self.l_dca = decline_curve()

    def determine_solve(self):
        # Use match/case to handle different input cases and calculate the missing variables

        match (None, None):
            case (self.qi, self.t_max):
                self.variables_to_solve = ['qi']
                self.qi = self.qf + self.de * self.eur
                self.t_max = 1200
            case (self.qi, self.qf):
                self.variables_to_solve = ['qi']
                self.qi = self.de * self.eur/2
                self.qf = 1
            case (self.qi, self.de):
                self.variables_to_solve = ['qi','de']
                self.qi = self.qf + self.dmin * self.eur
                self.de = (self.qi - self.qf) / self.eur
            case (self.qi, self.eur):
                self.variables_to_solve = ['qi','eur']
                self.qi = self.qf /self.de 
                self.eur = (self.qi - self.qf) / self.de
            case (self.t_max, self.qf):
                self.variables_to_solve = ['qf']
                self.qf = max(self.qi - self.de * self.eur,1)
                self.t_max = 1200
            case (self.t_max, self.de):
                self.variables_to_solve = ['de']
                self.de = (self.qi - self.qf) / self.eur
                self.t_max = 1200
            case (self.t_max, self.eur):
                self.variables_to_solve = ['eur']
                self.t_max = 1200
                self.eur = (self.qi - self.qf) / self.de
            case (self.qf, self.de):
                self.variables_to_solve = ['de']
                self.de = (self.qi) / self.eur
                self.qf = 1
            case (self.qf, self.eur):
                self.variables_to_solve = ['eur']
                self.eur = (self.qi) / self.de
                self.qf = 1
            case (self.de, self.eur):
                self.variables_to_solve = ['de','eur']
                self.eur = self.qi*self.t_max
                self.de = (self.qi - self.qf) / self.eur


    def dca_delta(self,vars_to_solve):

        for var_name, var_value in zip(self.variables_to_solve, vars_to_solve):
            setattr(self, var_name, var_value)

        self.l_dca.D_MIN = self.dmin
        t_range = np.array(range(0,int(self.t_max)))

        dca_array = np.array(self.l_dca.arps_decline(t_range,self.qi,self.de,self.b,0))

        


        dca_array = np.where(dca_array>self.qf,dca_array,0)


        self.l_t_max = len(np.where(dca_array > 0)[0])
        if self.l_t_max >0:
            self.l_qf = dca_array[np.where(dca_array > 0)[0][-1]]

        delta = np.sum(dca_array) - self.eur

        self.delta = delta

        return [delta] * len(self.variables_to_solve)
    
    def solve(self):

        self.determine_solve()

        initial_guess = [getattr(self, var) for var in self.variables_to_solve if getattr(self, var) is not None]

        result, infodict, ier, msg = fsolve(self.dca_delta, initial_guess, full_output=True)

        if ier==1:
            warning_flag = 0
        else:
            warning_flag = 1
        

        for var_name, var_value in zip(self.variables_to_solve, result):
            setattr(self, var_name, var_value)

        if 't_max' in self.variables_to_solve or len(self.variables_to_solve)==1:
            self.t_max = self.l_t_max

        if 'qf' in self.variables_to_solve or len(self.variables_to_solve)==1:
            self.qf = self.l_qf
        return self.qi, self.t_max, self.qf, self.de, self.eur, warning_flag, self.delta


class decline_curve:

    def __init__(self):
        #Constants
        self.DAY_NORM = 30.4
        self.GAS_CUTOFF = 3.2 #GOR for classifying well as gas or oil, MSCF/STB
        self.STAT_FILE = sys.stdout
        self.MINOR_TAIL = 6 #Number of months from tail to use for minor phase ratios
        self.SET_LENGTH = 5280 #Length to normalize horizontals to
        self.D_MIN = .08/12 #Minimum monthly decline rate
        self.DEBUG_ON = False
        #Settable 
        self.verbose = True
        self.FILTER_BONFP = .5 #Normally set to .5
        self.DEFAULT_DI  = .8/12
        self.DEFAULT_B = .5
        self.V_DCA_FAILURES = 0
        self.OUTLIER_CORRECTION = True
        self.IQR_LIMIT = 1.5
        self._min_h_b = .99
        self._max_h_b = 2
        
        self._backup_decline = False
        self._dataframe = None
        self._date_col = None
        self._phase_col = None
        self._length_col = None
        self._uid_col = None
        self._dayson_col = None
        self._oil_col = None
        self._gas_col = None
        self._water_col = None
        self._input_monthly = True

        self._force_t0 = False

        #Get only variables
        self._normalized_dataframe = pd.DataFrame()
        self._params_dataframe = None
        self._flowstream_dataframe = None
        self._typecurve = None
        self._oneline = pd.DataFrame()

        self.tc_params = pd.DataFrame()
        self.dca_param_df = []
        

    @property
    def dataframe(self):
        return self._dataframe


    @dataframe.setter
    def dataframe(self,value):
        self._dataframe = value

    @property
    def input_monthly(self):
        return self._input_monthly


    @input_monthly.setter
    def input_monthly(self,value):
        self._input_monthly = value

    @property
    def date_col(self):
        return self._date_col


    @date_col.setter
    def date_col(self,value):
        self._date_col = value

    @property
    def phase_col(self):
        return self._phase_col


    @phase_col.setter
    def phase_col(self,value):
        self._phase_col = value

    @property
    def length_col(self):
        return self._length_col


    @length_col.setter
    def length_col(self,value):
        self._length_col = value

    @property
    def uid_col(self):
        return self._uid_col


    @uid_col.setter
    def uid_col(self,value):
        self._uid_col = value

    @property
    def dayson_col(self):
        return self._dayson_col


    @dayson_col.setter
    def dayson_col(self,value):
        self._dayson_col = value

    @property
    def oil_col(self):
        return self._oil_col


    @oil_col.setter
    def oil_col(self,value):
        self._oil_col = value

    @property
    def gas_col(self):
        return self._gas_col


    @gas_col.setter
    def gas_col(self,value):
        self._gas_col = value

    @property
    def water_col(self):
        return self._water_col


    @water_col.setter
    def water_col(self,value):
        self._water_col = value


    @property
    def backup_decline(self):
        return self._backup_decline


    @backup_decline.setter
    def backup_decline(self,value):
        self._backup_decline = value

    @property
    def min_h_b(self):
        return self._min_h_b


    @min_h_b.setter
    def min_h_b(self,value):
        self._min_h_b= value


    @property
    def max_h_b(self):
        return self._max_h_b


    @max_h_b.setter
    def max_h_b(self,value):
        self._max_h_b= value

    @property
    def params_dataframe(self):
        return self._params_dataframe

    @property
    def flowstream_dataframe(self):
        return self._flowstream_dataframe

    @property
    def oneline_dataframe(self):
        return self._oneline

    @property
    def typecurve(self):
        return self._typecurve

    def month_diff(self, a, b):
        return 12 * (a.dt.year - b.dt.year) + (a.dt.month - b.dt.month)

    def day_diff(self,a,b):
        return (a - b) / np.timedelta64(1, 'D')

    def infill_production(self):
        """
        An error was found where gaps in the historical production would be infilled
        with the wrong P_DATE
        """

    def generate_t_index(self):
        #print(self._date_col, file=self.STAT_FILE, flush=True)
        self._dataframe[self._date_col] = pd.to_datetime(self._dataframe[self._date_col])
        min_by_well = self._dataframe[[self._uid_col,self._date_col]].groupby(by=[self._uid_col]).min().reset_index()
        min_by_well = min_by_well.rename(columns={self._date_col:'MIN_DATE'})
        #print(min_by_well)
        
        self._dataframe = self._dataframe.merge(
            min_by_well, 
            left_on = self._uid_col,
            right_on = self._uid_col,
            suffixes=(None,'_MIN')
        )

        if self._input_monthly:
            self._dataframe['T_INDEX'] = self.month_diff(
                self._dataframe[self._date_col],
                self._dataframe['MIN_DATE']
            )
        else:
            self._dataframe['T_INDEX'] = self.day_diff(
                self._dataframe[self._date_col],
                self._dataframe['MIN_DATE']
            )

        #return 0

    def assign_major(self):
        l_cum = self._normalized_dataframe[['UID','NORMALIZED_OIL','NORMALIZED_GAS']].groupby(by=['UID']).sum().reset_index()
        l_cum['MAJOR'] = np.where(
            l_cum["NORMALIZED_OIL"] >0,
            np.where(
                l_cum["NORMALIZED_GAS"]/l_cum['NORMALIZED_OIL']>self.GAS_CUTOFF,
                'GAS',
                'OIL'
            ),
            "GAS"
        )

        self._normalized_dataframe = self._normalized_dataframe.merge(
            l_cum,
            left_on = "UID",
            right_on = "UID",
            suffixes=(None,'_right')
        )

    def normalize_production(self):

        self._normalized_dataframe['UID'] = self._dataframe[self._uid_col]
        self._normalized_dataframe['T_INDEX'] = self._dataframe['T_INDEX']

        if self._length_col == None:
            self._normalized_dataframe['LENGTH_NORM'] = 1.0
        else:
            self._dataframe[self._length_col] = self._dataframe[self._length_col].fillna(0)

            self._normalized_dataframe['LENGTH_NORM'] = np.where(
                self._dataframe[self._length_col] > 1,
                self._dataframe[self._length_col],
                1
            )

        self._normalized_dataframe['HOLE_DIRECTION'] = np.where(
            self._normalized_dataframe['LENGTH_NORM']> 1,
            "H",
            "V"
        )

        if self._length_col == None:
            self._normalized_dataframe['LENGTH_SET'] = 1.0
        else:
            self._normalized_dataframe['LENGTH_SET'] = np.where(
                self._dataframe[self._length_col] > 1,
                self.SET_LENGTH,
                1.0
            )

        

        if self._dayson_col == None:
            self._normalized_dataframe['DAYSON'] = 30.4
        else:
            self._dataframe[self._dayson_col] = self._dataframe[self._dayson_col].fillna(30.4)

            self._normalized_dataframe['DAYSON'] = np.where(
                self._dataframe[self._dayson_col] > 0,
                self._dataframe[self._dayson_col],
                0
            )

        self._dataframe[self._oil_col] = pd.to_numeric(self._dataframe[self._oil_col], errors='coerce')
        self._dataframe[self._oil_col] = self._dataframe[self._oil_col].fillna(0)

        self._dataframe[self._gas_col] = pd.to_numeric(self._dataframe[self._gas_col], errors='coerce')
        self._dataframe[self._gas_col] = self._dataframe[self._gas_col].fillna(0)

        self._dataframe[self._water_col] = pd.to_numeric(self._dataframe[self._water_col], errors='coerce')
        self._dataframe[self._water_col] = self._dataframe[self._water_col].fillna(0)

        #self._normalized_dataframe.to_csv('outputs/test.csv')

        self._normalized_dataframe['NORMALIZED_OIL'] = (
            self._dataframe[self._oil_col]*
            self.DAY_NORM*
            self._normalized_dataframe['LENGTH_SET'] /
            (self._normalized_dataframe['LENGTH_NORM'] * self._normalized_dataframe['DAYSON'])
        )

        self._normalized_dataframe['NORMALIZED_GAS'] = (
            self._dataframe[self._gas_col]*
            self.DAY_NORM*
            self._normalized_dataframe['LENGTH_SET'] /
            (self._normalized_dataframe['LENGTH_NORM'] * self._normalized_dataframe['DAYSON'])
        )

        self._normalized_dataframe['NORMALIZED_WATER'] = (
            self._dataframe[self._water_col]*
            self.DAY_NORM*
            self._normalized_dataframe['LENGTH_SET'] /
            (self._normalized_dataframe['LENGTH_NORM'] * self._normalized_dataframe['DAYSON'])
        )

        
        if self._phase_col == None:
            self.assign_major()
        else:
            self._normalized_dataframe['MAJOR'] = self._dataframe[self._phase_col]
        

        self._normalized_dataframe = self._normalized_dataframe[[
            'UID',
            'LENGTH_NORM',
            "HOLE_DIRECTION",
            'MAJOR',
            'T_INDEX',
            'NORMALIZED_OIL',
            'NORMALIZED_GAS',
            'NORMALIZED_WATER'
        ]]

        self._normalized_dataframe['NORMALIZED_OIL'] = self._normalized_dataframe['NORMALIZED_OIL'].fillna(0) 
        self._normalized_dataframe['NORMALIZED_GAS'] = self._normalized_dataframe['NORMALIZED_GAS'].fillna(0) 
        self._normalized_dataframe['NORMALIZED_WATER'] = self._normalized_dataframe['NORMALIZED_WATER'].fillna(0) 
    
        if self.DEBUG_ON:
            self._normalized_dataframe.to_csv('outputs/norm_test.csv')
    
    def outlier_detection(self, input_x, input_y):

        
        filtered_x = []
        filtered_y = []
    
        ln_input_y= np.log(input_y)

        if len([i for i in ln_input_y if i > 0]) > 0:
            
            regression = sm.formula.ols("data ~ x", data=dict(data=ln_input_y, x=input_x)).fit()
            try:
                test = regression.outlier_test()
                
                for index, row in test.iterrows():
                    if row['bonf(p)']> self.FILTER_BONFP:
                        filtered_x.append(input_x[index])
                        filtered_y.append(input_y[index])
            except:
                if self.verbose:
                    print('Error in outlier detection.')
                filtered_x = input_x
                filtered_y = input_y

        return filtered_x, filtered_y

    def arps_decline(self,x,qi,di,b,t0):
        if qi > 0 and not math.isinf(qi):
            problemX = t0-1/(b*di)
            #print(di,self.D_MIN,b,qi)
            if di < self.D_MIN:
                qlim = qi
                di = self.D_MIN
                tlim = -1
            else:
                qlim = qi*(self.D_MIN/di)**(1/b)
                #print(qlim)
                try:
                    tlim = int(((qi/qlim)**(b)-1)/(b*di)+t0)
                    #q_at_lim = (qi)/(1+b*(di)*(int(tlim)-t0))**(1/b)
                except:
                    #tlim=-1
                    print(qi,qlim,di,b)
            #problemX = t0+1
            #print(tlim)
            try:
                q_x = np.where(
                    x>problemX,
                    np.where(x<tlim,
                        (qi)/(1+b*(di)*(x-t0))**(1/b),
                        qlim*np.exp(-self.D_MIN*(x-tlim))
                    ),
                    0
                )
            except Exception as e:
                print(qi,qlim,di,b)
                raise e
            #print(q_x)
            #qi = (qi)/(1+b*(ai)*(x))**(1/b)
        else:
            q_x = [0.0 for _ in x]
        return q_x
    
    def handle_dca_error(self,s,x_vals,y_vals):
        if s["MAJOR"] == 'OIL':
            #print(sum_df)
            minor_ratio = np.sum(s['NORMALIZED_GAS'][-self.MINOR_TAIL:])/np.sum(s['NORMALIZED_OIL'][-self.MINOR_TAIL:])
            water_ratio = np.sum(s['NORMALIZED_WATER'][-self.MINOR_TAIL:])/np.sum(s['NORMALIZED_OIL'][-self.MINOR_TAIL:])
        else:
            minor_ratio = np.sum(s['NORMALIZED_OIL'][-self.MINOR_TAIL:])/np.sum(s['NORMALIZED_GAS'][-self.MINOR_TAIL:])
            water_ratio = np.sum(s['NORMALIZED_WATER'][-self.MINOR_TAIL:])/np.sum(s['NORMALIZED_GAS'][-self.MINOR_TAIL:])
        i = -1
        while i > -len(x_vals):
            if y_vals[i]>0:
                break
            else:
                i -= 1
        s['qi']=y_vals[i]
        s['di']=self.DEFAULT_DI
        s['b']=self.DEFAULT_B
        s['t0']=x_vals[i]
        s['q0']=y_vals[0] #Probably will need revision, high chance first value is zero
        s['minor_ratio']=minor_ratio
        s['water_ratio']=water_ratio

        return s

    def dca_params(self,s):

        x_vals = s['T_INDEX']

        if s['MAJOR'] == 'OIL':
            y_vals = s['NORMALIZED_OIL']
        else:
            y_vals = s['NORMALIZED_GAS']

        if len(x_vals) > 3:
            z = np.array(y_vals)
            a = argrelextrema(z, np.greater)
            if len(a[0]) > 0:
                indexMax = a[-1][-1]
                indexMin = a[-1][0]
                t0Max = x_vals[indexMax]
                t0Min = x_vals[indexMin]
            else:
                indexMax = 0
                indexMin = 0
                t0Max = x_vals[indexMax]
                t0Min = x_vals[indexMin]
            

            filtered_x = np.array(x_vals[indexMin:])
            filtered_y = np.array(y_vals[indexMin:])

            zero_filter = np.array([y > 0 for y in filtered_y])
            filtered_x = filtered_x[zero_filter]
            filtered_y = filtered_y[zero_filter]
            
            outliered_x, outliered_y = self.outlier_detection(filtered_x,filtered_y)

            if self._force_t0:
                outliered_x = x_vals
                outliered_y = y_vals

            

            if len(outliered_x) > 3:
                if t0Min == t0Max:
                    t0Max = t0Max + 1
                try:
                    di_int = np.log(outliered_y[0]/outliered_y[-1])/(outliered_x[-1]-outliered_x[0])
                except ZeroDivisionError:
                    di_int = .1
                except Exception as e:
                    raise(e)
                q_max = np.max(outliered_y)
                q_min = np.min(outliered_y)

                if s['HOLE_DIRECTION'] == 'H':
                    bMin = self._min_h_b
                    bMax = self._max_h_b
                else:
                    bMin = self._min_h_b
                    bMax = self._max_h_b

                if di_int < 0:
                    di_int = np.log(q_max/q_min)/(outliered_x[outliered_y.index(q_min)]-outliered_x[outliered_y.index(q_max)])
                
                if di_int < 0:
                    if q_max == outliered_y[-1]:
                        di_int = .1
                    else:
                        di_int = np.log(q_max/outliered_y[-1])/(outliered_x[-1]-outliered_x[outliered_y.index(q_max)])
                
                if self._force_t0:
                    weight_range = [1 for _ in range(1,len(outliered_x)+1)]
                    #weight_range = list(range(1,len(outliered_x)+1))
                    di_min = .01
                    di_max = .9
                    t0Min = 1
                    t0Max = 2
                else:
                    di_min = di_int/2
                    di_max = di_int*2
                    weight_range = list(range(1,len(outliered_x)+1))
                    weight_range = weight_range[::-1]
                
                try:
                    popt, pcov = curve_fit(self.arps_decline, outliered_x, outliered_y,
                        p0=[q_max, di_int,(bMin+bMax)/2,t0Min], 
                        bounds=([q_min,di_min,bMin, t0Min], [q_max*1.1,di_max,bMax,t0Max]),
                        sigma = weight_range, absolute_sigma = True)
                    
                    

                    if s["MAJOR"] == 'OIL':
                        #print(sum_df)
                        minor_ratio = np.sum(s['NORMALIZED_GAS'][-self.MINOR_TAIL:])/np.sum(s['NORMALIZED_OIL'][-self.MINOR_TAIL:])
                        if self.verbose:
                            print(f'Minor Ratio: {minor_ratio}')
                        water_ratio = np.sum(s['NORMALIZED_WATER'][-self.MINOR_TAIL:])/np.sum(s['NORMALIZED_OIL'][-self.MINOR_TAIL:])
                        if self.verbose:
                            print(f'Water Ratio: {water_ratio}')
                    else:
                        minor_ratio = np.sum(s['NORMALIZED_OIL'][-self.MINOR_TAIL:])/np.sum(s['NORMALIZED_GAS'][-self.MINOR_TAIL:])
                        water_ratio = np.sum(s['NORMALIZED_WATER'][-self.MINOR_TAIL:])/np.sum(s['NORMALIZED_GAS'][-self.MINOR_TAIL:])

                    if not math.isinf(popt[0]):

                        s['qi']=popt[0]
                        s['di']=popt[1]
                        s['b']=popt[2]
                        s['t0']=popt[3]
                        s['q0']=y_vals[0] #Probably will need revision, high chance first value is zero
                        s['minor_ratio']=minor_ratio
                        s['water_ratio']=water_ratio
                    else:
                        self.V_DCA_FAILURES += 1
                        if self.verbose:
                            print('DCA Error: '+str(s['UID']), file=self.STAT_FILE, flush=True)

                        if self._backup_decline:
                            #local_df = self._normalized_dataframe.loc[self._normalized_dataframe['UID'] == w]
                            #sum_df = local_df.tail(self.MINOR_TAIL).sum()
                            return self.handle_dca_error(s,x_vals, y_vals)
                except:
                    self.V_DCA_FAILURES += 1
                    if self.verbose:
                        print('DCA Error: '+str(s['UID']), file=self.STAT_FILE, flush=True)

                    if self._backup_decline:
                        #local_df = self._normalized_dataframe.loc[self._normalized_dataframe['UID'] == w]
                        #sum_df = local_df.tail(self.MINOR_TAIL).sum()
                        return self.handle_dca_error(s,x_vals, y_vals)
            else:
                self.V_DCA_FAILURES += 1
                if self.verbose:
                    print('Base x: {}  Filtered x: {}  Outliered x: {}'.format(len(x_vals),len(filtered_x),len(outliered_x)))
                    print('Insufficent data after filtering, well: '+str(s['UID']), file=self.STAT_FILE, flush=True)
                if self._backup_decline:
                    return self.handle_dca_error(s,x_vals, y_vals)

        else :
            self.V_DCA_FAILURES += 1
            if self.verbose:
                print('Insufficent data before filtering, well: '+str(s['UID']), file=self.STAT_FILE, flush=True)
            if self._backup_decline:
                return self.handle_dca_error(s,x_vals, y_vals)

        #print(s)
        return s
    
    def vect_generate_params_tc(self,param_df):

        self._force_t0 = True

        param_df['HOLE_DIRECTION'] = "H"
        param_df = param_df[param_df['T_INDEX']<60]
        param_df = param_df.rename(columns={
            'OIL':'NORMALIZED_OIL',
            'GAS':"NORMALIZED_GAS",
            'WATER':'NORMALIZED_WATER',
            'level_1':'UID'
        })

        imploded_df = param_df[[
            'UID',
            'MAJOR',
            'HOLE_DIRECTION',
            'T_INDEX',
            'NORMALIZED_OIL',
            'NORMALIZED_GAS',
            'NORMALIZED_WATER'
        ]].groupby(
            ['UID',
            'MAJOR',
            'HOLE_DIRECTION']
        ).agg({
            'T_INDEX': lambda x: x.tolist(),
            'NORMALIZED_OIL': lambda x: x.tolist(),
            'NORMALIZED_GAS': lambda x: x.tolist(),
            'NORMALIZED_WATER': lambda x: x.tolist()
        }).reset_index()

        imploded_df = imploded_df.apply(self.dca_params, axis=1)
        imploded_df = imploded_df[[
            'UID',
            'MAJOR',
            'q0',
            'qi',
            'di',
            'b',
            't0',
            'minor_ratio',
            'water_ratio',
        ]].rename(columns={
            'MAJOR':'major',
        })

        self._force_t0 = False

        return imploded_df



    def vect_generate_params(self):
        self.V_DCA_FAILURES = 0
        l_start = time.time()

        imploded_df = self._normalized_dataframe[[
            'UID',
            'MAJOR',
            'HOLE_DIRECTION',
            'LENGTH_NORM',
            'T_INDEX',
            'NORMALIZED_OIL',
            'NORMALIZED_GAS',
            'NORMALIZED_WATER'
        ]].groupby(
            ['UID',
            'MAJOR',
            'HOLE_DIRECTION',
            'LENGTH_NORM']
        ).agg({
            'T_INDEX': lambda x: x.tolist(),
            'NORMALIZED_OIL': lambda x: x.tolist(),
            'NORMALIZED_GAS': lambda x: x.tolist(),
            'NORMALIZED_WATER': lambda x: x.tolist()
        }).reset_index()

        imploded_df = imploded_df.apply(self.dca_params, axis=1)

        imploded_df = imploded_df[[
            'UID',
            'MAJOR',
            'LENGTH_NORM',
            'q0',
            'qi',
            'di',
            'b',
            't0',
            'minor_ratio',
            'water_ratio',
        ]].rename(columns={
            'MAJOR':'major',
            'LENGTH_NORM':'h_length'
        })

        r_df:pd.DataFrame = pd.DataFrame([])

        for major in ['OIL','GAS']:
            l_df = imploded_df[imploded_df['major']==major]

            if len(l_df)>0:
                if self.OUTLIER_CORRECTION:
                    q3, q2, q1 = np.percentile(l_df['minor_ratio'], [75,50 ,25])
                    high_cutoff = self.IQR_LIMIT*(q3-q1)+q3
                    l_df['minor_ratio'] = np.where(
                        l_df['minor_ratio']>high_cutoff,
                        q2,
                        l_df['minor_ratio']
                    )

                    q3, q2, q1 = np.percentile(l_df['water_ratio'], [75,50 ,25])
                    high_cutoff = self.IQR_LIMIT*(q3-q1)+q3
                    l_df['water_ratio'] = np.where(
                        l_df['water_ratio']>high_cutoff,
                        q2,
                        l_df['water_ratio']
                    )

                if r_df.empty:
                    r_df = l_df
                else:
                    r_df = pd.concat([r_df,l_df])

        imploded_df = r_df

        print('Total DCA Failures: '+str(self.V_DCA_FAILURES), file=self.STAT_FILE, flush=True)
        print(f'Total wells analyzed: {len(imploded_df)}', file=self.STAT_FILE, flush=True)
        print('Failure rate: {:.2%}'.format(self.V_DCA_FAILURES/len(imploded_df)), file=self.STAT_FILE, flush=True)
        l_duration = time.time() - l_start
        print("Vectorized DCA generation: {:.2f} seconds".format(l_duration), file=self.STAT_FILE, flush=True)

        self._params_dataframe = imploded_df


    def run_DCA(self, _verbose=True):
        self.verbose = _verbose
        if self.verbose:
            print('Generating time index.', file=self.STAT_FILE, flush=True)
            
        
        self.generate_t_index()

        if self.verbose:
            print('Normalizing production.', file=self.STAT_FILE, flush=True)

        self.normalize_production()

        if self.verbose:
            print('Generating decline parameters.', file=self.STAT_FILE, flush=True)
        #self.generate_params()
        
        self.vect_generate_params()

    def add_months(self, start_date, delta_period):
        end_date = start_date + pd.DateOffset(months=delta_period)
        return end_date
    
    def generate_oneline(self, num_months=1200, denormalize=False, _verbose=False):
        self.verbose = _verbose

        self.generate_flowstream(num_months=num_months,denormalize=denormalize,actual_dates=False,_verbose=_verbose)

        if self._params_dataframe.empty:
            self.run_DCA(_verbose=_verbose)

        t_range = np.array(range(1,num_months))

        flow_dict = {
            'UID':[],
            'MAJOR':[],
            "IPO":[],
            'IPG':[],
            'B':[],
            'DE':[],
            'T0':[],
            'MINOR_RATIO':[],
            'WATER_RATIO':[]
        }

        #param_df = self.vect_generate_params_tc(self._flowstream_dataframe)

        online_df = self._flowstream_dataframe[['UID','OIL',"GAS",'WATER']].groupby('UID').sum().reset_index()

        #print(self._flowstream_dataframe)

        self._dataframe[self._date_col] = pd.to_datetime(self._dataframe[self._date_col])

        min_df = self._dataframe[[self._uid_col,self._date_col]].groupby(by=[self._uid_col]).min().reset_index()
        #min_df = self._flowstream_dataframe.groupby(['UID']).min().reset_index()
        min_df = min_df.rename(columns={self._uid_col:"UID",self._date_col:"MIN_DATE"})
        min_df = min_df[min_df['MIN_DATE'].notnull()]

        self._params_dataframe = self._params_dataframe.merge(min_df, left_on='UID', right_on='UID')

        self._params_dataframe = self._params_dataframe.replace([np.inf, -np.inf], np.nan)

        self._params_dataframe = self._params_dataframe.dropna(subset='t0')

        self._params_dataframe['T0_DATE'] =  self._params_dataframe.apply(lambda row: self.add_months(row["MIN_DATE"], round(row["t0"],0)), axis = 1)

        #print(self._params_dataframe)

        for index, row in self._params_dataframe.iterrows():
            if denormalize and row['h_length']>1:
                denormalization_scalar = row['h_length']/self.SET_LENGTH
            else:
                denormalization_scalar = 1
            
            dca = self._flowstream_dataframe[self._flowstream_dataframe['UID']==row['UID']]

            

            if np.sum(dca[row['major']]) > 0:

                flow_dict['UID'].append(row['UID'])
                flow_dict['MAJOR'].append(row['major'])
                
                flow_dict['B'].append(row['b'])
                flow_dict['DE'].append(row['di'])
                flow_dict['T0'].append(row['T0_DATE'])
                flow_dict['MINOR_RATIO'].append(row['minor_ratio'])
                flow_dict['WATER_RATIO'].append(row['water_ratio'])
                if row['major'] == "OIL":
                    #flow_dict['IPO'].append(max(dca[row['major']]))
                    flow_dict['IPO'].append(row['qi']*denormalization_scalar)
                    if np.isnan(row['minor_ratio']):
                        flow_dict['IPG'].append(max(dca[row['major']])*0)
                    else:
                        #flow_dict['IPG'].append(max(dca[row['major']])*row['minor_ratio'])
                        flow_dict['IPG'].append(row['qi']*row['minor_ratio'])
                else:
                    #flow_dict['IPG'].append(max(dca[row['major']]))
                    flow_dict['IPG'].append(row['qi']*denormalization_scalar)
                    if np.isnan(row['minor_ratio']):
                        flow_dict['IPO'].append(max(dca[row['major']])*0)
                    else:
                        #flow_dict['IPO'].append(max(dca[row['major']])*row['minor_ratio'])
                        flow_dict['IPO'].append(row['qi']*row['minor_ratio'])
        flow_dict = pd.DataFrame(flow_dict)
        online_df = online_df.merge(flow_dict,left_on='UID',right_on='UID')
        online_df['ARIES_DE'] = online_df.apply(lambda x: (1-np.power(((x.DE*12)*x.B+1),(-1/x.B)))*100, axis=1)
        self._oneline = online_df

    def generate_flowstream(self, num_months=1200, denormalize=False, actual_dates=False, _verbose=False):
        self.verbose = _verbose

        if self._params_dataframe == None:
            self.run_DCA(_verbose=_verbose)

        t_range = np.array(range(1,num_months))

        flow_dict = {
            'UID':[],
            'MAJOR':[],
            'T_INDEX':[],
            'OIL':[],
            'GAS':[],
            'WATER':[]
        }

        

        for index, row in self._params_dataframe.iterrows():
            if denormalize and row['h_length']>1:
                denormalization_scalar = row['h_length']/self.SET_LENGTH
            else:
                denormalization_scalar = 1
            
            dca = np.array(self.arps_decline(t_range,row.qi,row.di,row.b,row.t0))*denormalization_scalar
            if np.sum(dca) > 0:
                flow_dict['UID'].append(row['UID'])
                flow_dict['MAJOR'].append(row['major'])
                flow_dict['T_INDEX'].append(t_range)
                if row['major'] == "OIL":
                    flow_dict['OIL'].append(dca)
                    if np.isnan(row['minor_ratio']):
                        flow_dict['GAS'].append(dca*0)
                    else:
                        flow_dict['GAS'].append(dca*row['minor_ratio'])
                else:
                    flow_dict['GAS'].append(dca)
                    if np.isnan(row['minor_ratio']):
                        flow_dict['OIL'].append(dca*0)
                    else:
                        flow_dict['OIL'].append(dca*row['minor_ratio'])
                if np.isnan(row['water_ratio']):
                    flow_dict['WATER'].append(dca*0)
                else:
                    flow_dict['WATER'].append(dca*row['water_ratio'])


        self._flowstream_dataframe = pd.DataFrame(flow_dict)
        #print(self._flowstream_dataframe.columns)
        self._flowstream_dataframe = self._flowstream_dataframe.set_index(['UID','MAJOR']).apply(pd.Series.explode).reset_index()
        self._flowstream_dataframe = self._flowstream_dataframe.set_index(['UID', 'T_INDEX'])

        self._flowstream_dataframe['OIL'] = pd.to_numeric(
            self._flowstream_dataframe['OIL']
        )

        self._flowstream_dataframe['GAS'] = pd.to_numeric(
            self._flowstream_dataframe['GAS']
        )

        self._flowstream_dataframe['WATER'] = pd.to_numeric(
            self._flowstream_dataframe['WATER']
        )

        self._flowstream_dataframe.replace([np.inf, -np.inf], 0, inplace=True)

        if denormalize:
            actual_df = self._dataframe[[self._uid_col,'T_INDEX',self._oil_col,self._gas_col,self._water_col]]
            actual_df = actual_df.rename(columns={
                self._uid_col:'UID',
                self._oil_col:'OIL',
                self._gas_col:"GAS",
                self._water_col:"WATER"
            })
        else:
            actual_df = self._normalized_dataframe[[
                'UID',
                'T_INDEX',
                'NORMALIZED_OIL',
                'NORMALIZED_GAS',
                'NORMALIZED_WATER'
            ]]
            actual_df = actual_df.rename(columns={
                'NORMALIZED_OIL':'OIL',
                'NORMALIZED_GAS':"GAS",
                'NORMALIZED_WATER':'WATER'
            })

        if actual_dates:
            actual_df['P_DATE'] = self._dataframe[self._date_col]
            self._flowstream_dataframe['P_DATE'] = None
            
        # Added to ensure UID and T_INDEX are unique
    
        actual_df = actual_df.set_index(['UID', 'T_INDEX'])

        #actual_df.to_csv('outputs/actual.csv')
        #self._flowstream_dataframe.to_csv('outputs/flowstrems.csv')

        #self._flowstream_dataframe.to_csv('outputs/flowstream_df.csv')
        #actual_df.to_csv('outputs/actual.csv')
        actual_df = actual_df[~actual_df.index.duplicated(keep='first')]
        self._flowstream_dataframe = self._flowstream_dataframe[~self._flowstream_dataframe.index.duplicated(keep='first')]
        self._flowstream_dataframe.update(actual_df)
        self._flowstream_dataframe = self._flowstream_dataframe.reset_index()


        if actual_dates:
            
            # Updated code using the fact that T_INDEX is always referenced to MIN_DATE

            self._flowstream_dataframe['P_DATE'] = pd.to_datetime(self._flowstream_dataframe['P_DATE'])

            min_df = self._dataframe[[self._uid_col,self._date_col]].groupby(by=[self._uid_col]).min().reset_index()
            #min_df = self._flowstream_dataframe.groupby(['UID']).min().reset_index()
            min_df = min_df.rename(columns={self._uid_col:"UID",self._date_col:"MIN_DATE"})
            min_df = min_df[min_df['MIN_DATE'].notnull()]

            self._flowstream_dataframe = self._flowstream_dataframe.merge(min_df, left_on='UID', right_on='UID')

            self._flowstream_dataframe = self._flowstream_dataframe.replace([np.inf, -np.inf], np.nan)

            self._flowstream_dataframe['P_DATE'] = np.where(
                self._flowstream_dataframe['P_DATE'].isnull(),
                self._flowstream_dataframe.apply(lambda row: self.add_months(row["MIN_DATE"], row["T_INDEX"]), axis = 1),
                self._flowstream_dataframe['P_DATE']
            )

            self._flowstream_dataframe = self._flowstream_dataframe.drop(['MIN_DATE'],axis=1)

            # Orginal code encountered issues when date skips in historical were present

            #self._flowstream_dataframe['P_DATE'] = pd.to_datetime(self._flowstream_dataframe['P_DATE'])
            
            #cum_count = self._flowstream_dataframe[self._flowstream_dataframe['P_DATE'].isnull()].groupby(['UID']).cumcount().rename('OFFSET_INDEX')
            #cum_count = cum_count+1
            #self._flowstream_dataframe = self._flowstream_dataframe.merge(cum_count,how='left', left_index=True, right_index=True)
            
            #max_df = self._flowstream_dataframe.groupby(['UID']).max().reset_index()
            #max_df = max_df[['UID','P_DATE']].rename(columns={'P_DATE':'MAX_DATE'})
            #max_df = max_df[max_df['MAX_DATE'].notnull()]
            
            #self._flowstream_dataframe = self._flowstream_dataframe.merge(max_df, left_on='UID', right_on='UID')
            
            #self._flowstream_dataframe = self._flowstream_dataframe.replace([np.inf, -np.inf], np.nan)
            
            #self._flowstream_dataframe['OFFSET_INDEX'] = self._flowstream_dataframe['OFFSET_INDEX'].fillna(0)

            #self._flowstream_dataframe['P_DATE'] = np.where(
            #    self._flowstream_dataframe['P_DATE'].isnull(),
            #    self._flowstream_dataframe.apply(lambda row: self.add_months(row["MAX_DATE"], row["OFFSET_INDEX"]), axis = 1),
            #    self._flowstream_dataframe['P_DATE']
            #)

            #self._flowstream_dataframe = self._flowstream_dataframe.drop(['OFFSET_INDEX','MAX_DATE'],axis=1)

    def generate_typecurve(self, num_months=1200, denormalize=False, prob_levels=[.1,.5,.9], _verbose=False, return_params=False):
        if self._flowstream_dataframe == None:
            self.generate_flowstream(num_months=num_months,denormalize=denormalize, _verbose=_verbose)

        return_df = self._flowstream_dataframe.reset_index()
        
        #print(return_df.head())
        #return_df = self._flowstream_dataframe[['T_INDEX','OIL','GAS','WATER']]
        if self.DEBUG_ON:
            return_df.to_csv('outputs/test_quantiles.csv')
        
        #print(self._flowstream_dataframe)

        return_df = self._flowstream_dataframe[['T_INDEX','OIL','GAS','WATER']].groupby(['T_INDEX']).quantile(prob_levels).reset_index()
        avg_df = self._flowstream_dataframe[['T_INDEX','OIL','GAS','WATER']].groupby(['T_INDEX']).mean().reset_index()
        avg_df['level_1'] = 'mean'
        return_df = pd.concat([return_df,avg_df])
        #vect_df = self._flowstream_dataframe[['T_INDEX','MAJOR','OIL','GAS','WATER']].groupby(['T_INDEX','MAJOR']).quantile(prob_levels).reset_index()

        
        if return_params:
            r_df = pd.DataFrame([])
            for major in ['OIL','GAS']:
            #return_df.to_csv('outputs/return_test.csv')
                l_df = return_df.copy()
                l_df['MAJOR'] = major
                param_df = self.vect_generate_params_tc(l_df)
                param_df['d0'] = param_df.apply(lambda x: x.di*np.power((1+x.b*x.di*(1-x.t0)),-1), axis=1)
                #param_df['d0_a'] = param_df.apply(lambda x: np.power((1+x.d0),12)-1, axis=1)
                param_df['d0_a'] = param_df.apply(lambda x: x.d0*12, axis=1)
                param_df['aries_de'] = param_df.apply(lambda x: (1-np.power((x.d0_a*x.b+1),(-1/x.b)))*100, axis=1)
                param_df = param_df.rename(columns={
                    'qi':'Actual Initial Rate, bbl/month',
                    'q0':'DCA Initial Rate, bbl/month',
                    'di':'Nominal Initial Decline at Match Point, fraction/months',
                    'b':'B Factor, unitless',
                    't0':'Match Point, months',
                    'minor_ratio':'Minor Phase Ratio, (M/B or B/M)',
                    'water_ratio':'Water Phase Ratio (B/B or B/M)',
                    'd0':'Nominal Initial Decline at Time Zero, fraction/months',
                    'd0_a':'Nominal Initial Decline at Time Zero, fraction/years',
                    'aries_de':'Effective Initial Decline at Time Zero, %/years (FOR ARIES)',
                    'UID':'Probability',
                    'major':'Major Phase'
                })
                if r_df.empty:
                    r_df = param_df
                else:
                    r_df = pd.concat([r_df,param_df])
            self.tc_params = r_df
        return_df = return_df.pivot(
                index=['T_INDEX'],
                columns='level_1',
                values=['OIL','GAS','WATER']
            )
            #oil_df.columns = ['P10 Oil, bbl/(km-month)','P50 Oil, bbl/(km-month)','P90 Oil, bbl/(km-month)']
        #oil_df = oil_df.rename(columns={'T_INDEX':'Months Online'})

        self._typecurve = return_df


if __name__ == '__main__':

    l_dca = decline_solver(
        qi=16805,
        qf=3000,
        eur=1104336.17516371,
        b=.01,
        dmin=.01/12
    )

    print(l_dca.solve())



#https://towardsdatascience.com/how-to-package-your-python-code-df5a7739ab2e
#https://packaging.python.org/tutorials/packaging-projects/

# Use the following commands to build
# python setup.py bdist_conda
# conda install --use-local resaid

# For pip
# pip install -e .

import setuptools


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="resaid",
    version="0.1.6",
    author="Greg Easley",
    author_email="greg@easley.dev",
    description="Reservoir engineering tools",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gregeasley/resaid",
    project_urls={
        "Bug Tracker": "https://github.com/gregeasley/resaid/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy>=1.22',
        'pandas>=1.5.3',
        'scipy>=1.0.0',
        'statsmodels>=0.13.5',
        'tqdm>=4.65.0',
    ],
    packages=setuptools.find_packages(),
    #python_requires=">=3.6",
)